// Copyright 2025 "Johan Maasing" <johan@zoom.nu>
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
package nu.zoom.dsl.cli;

import nu.zoom.dsl.ast.DocumentNode;
import nu.zoom.dsl.ast.ParserWrapper;
import nu.zoom.dsl.freemarker.Generator;
import picocli.CommandLine;
import picocli.CommandLine.Command;
import picocli.CommandLine.Option;
import picocli.CommandLine.Parameters;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.concurrent.Callable;

@Command(
	name = "run.sh",
	mixinStandardHelpOptions = true,
	description = "Generate source code from an endpoints specification file."
)
public class EndpointsCLI implements Callable<Integer> {
	public enum ParserType {
		Endpoints,
		States
	}
	@SuppressWarnings("unused")
    @Parameters(index = "0", description = "The source endpoints DSL file.")
	private Path file;

	@SuppressWarnings("CanBeFinal")
    @Option(names = {"-t", "--template"}, defaultValue = "endpoints-template", description = "The template directory. Default is ${DEFAULT-VALUE}")
	private Path templateDir ;

	@SuppressWarnings("CanBeFinal")
    @Option(names = {"-o", "--output"}, defaultValue = "endpoints-output", description = "The directory to write the generated code to. Default is ${DEFAULT-VALUE}")
	private Path outputDir ;

	@SuppressWarnings("unused")
    @Option(names = {"-v", "--verbose"}, description = "Print verbose debug messages.")
	private Boolean verbose = false;

	@Option(names = {"-p", "--parser"}, description = "Force use of a specific parser instead of determining from filename. Valid values: ${COMPLETION-CANDIDATES}.")
	private ParserType parser = null;

	public static void main(String[] args) {
		int exitCode = new CommandLine(new EndpointsCLI()).execute(args);
		System.exit(exitCode);
	}

	@Override
	public Integer call() {
		try {
			validateTemplateDirectory();
			validateInputFile();
			validateOutputDirectory();
			verbose("Parsing: " + file.toAbsolutePath());
			if (parser == null) {
				if (file.getFileName().toString().endsWith(".states")) {
					parser = ParserType.States;
				}
			}
			final DocumentNode rootNode ;
			if (parser == ParserType.States) {
				verbose("using state grammar.") ;
				rootNode = ParserWrapper.parseStates(file);
			} else {
				verbose("using endpoints grammar.") ;
				rootNode = ParserWrapper.parseEndpoints(file);
			}
			verbose("AST: " + rootNode);
			verbose("Generating from templates in: " + templateDir.toAbsolutePath());
			Generator generator = new Generator(templateDir, rootNode, outputDir);
			List<Path> generatedPaths = generator.generate();
			if (generatedPaths.isEmpty()) {
				System.out.println("No generated paths found.");
			} else {
				generatedPaths.forEach(p -> verbose("Generated: " + p.toAbsolutePath()));
			}
			return 0;
		} catch (Exception e) {
			System.err.println(e.getMessage());
			return 1;
		}
	}

	private void verbose(String message) {
		if (this.verbose) {
			System.out.println(message);
		}
	}

	private void validateOutputDirectory() throws IOException {
		if (Files.notExists(this.outputDir)) {
			Files.createDirectories(this.outputDir);
		}
		if (!Files.isDirectory(this.outputDir)) {
			throw new IllegalArgumentException("Output directory: '" + this.outputDir + " 'is not a directory.");
		}
	}

	private void validateTemplateDirectory() {
		if (!Files.isDirectory(this.templateDir)) {
			throw new IllegalArgumentException("Template directory '" + this.templateDir + "' is not a directory.");
		}
	}

	private void validateInputFile() {
		if (Files.notExists(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' does not exist.");
		}
		if (!Files.isReadable(this.file) || !Files.isRegularFile(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' is not a readable file.");
		}
	}
}
